using System;
using UnityEngine;
using System.Collections;
using System.Linq;
using System.Collections.Generic;

namespace SPQR
{

    public class AEState
    {
        MVRScript X;
        public string name;

        float valueCurrent = 0f;
        public float valueMin = 0f;
        public float valueMax = 1f;

        float timeCurrent = 0f;
        public float timeMin = 0f;
        public float timeMax = 0f;
        float timeNext = 0f;

        bool delay = false;
        public float delayMin = 0f;
        public float delayMax = 0f;

        Func<bool> callback;


        public Dictionary<string, string> strings = new Dictionary<string, string>();
        public Dictionary<string, float> floats = new Dictionary<string, float>();
        public Dictionary<string, int> ints = new Dictionary<string, int>();
        public Dictionary<string, bool> bools = new Dictionary<string, bool>();

        public AEState(MVRScript script, string name, float valmin, float valmax, float timemin, float delaymin, float delaymax, float timemax, Func<bool> f = null)
        {
            X = script;
            valueMin = valmin;
            valueMax = valmax;
            timeMin = timemin;
            timeMax = timemax;
            this.name = name;
            this.delayMin = delaymin;
            this.delayMax = delaymax;
            callback = f;
            Randomize();
        }

        public AEState(MVRScript script, string name, Vector2 val)
        {
            X = script;
            valueMin = val.x;
            valueMax = val.y;
            timeMin = 0f;
            timeMax = 0f;
            this.name = name;
            this.delayMin = 0f;
            this.delayMax = 0f;
            callback = null;
            Randomize();
        }
        public AEState(MVRScript script, string name, Vector2 val, Vector2 time, Vector2 delay, Func<bool> f = null)
        {
            X = script;
            valueMin = val.x;
            valueMax = val.y;
            timeMin = time.x;
            timeMax = time.y;
            this.name = name;
            this.delayMin = delay.x;
            this.delayMax = delay.y;
            callback = f;
            Randomize();
        }

        public float GetValue()
        {
            Update();
            return valueCurrent;
        }
        public float GetTimeDuration()
        {
            return timeCurrent;
        }
        public float GetTimeStarted()
        {
            return timeNext - timeCurrent;
        }
        public float GetTimeElapsed()
        {
            return Time.time - GetTimeStarted();
        }
        public float GetTimeRemaining()
        {
            return timeNext - Time.time;
        }
        public float GetT()
        {
            return GetTimeElapsed() / GetTimeDuration();
        }
        public bool IsT1()
        {
            return GetT() <= 0.5f;
        }
        public float GetT1()
        {
            return GetT() * 2;
        }
        public float GetT2()
        {
            return (GetT1() - 0.5f) * 2;
        }

        public void Update()
        {
            if (IsExpired()) Randomize();
        }

        public bool IsExpired()
        {
            return Time.time > timeNext;
        }

        public bool IsDelayed()
        {
            return delay;
        }

        public void Randomize()
        {
            var valueCurrentOriginal = valueCurrent;
            valueCurrent = UnityEngine.Random.Range(valueMin, valueMax);
            timeCurrent = UnityEngine.Random.Range(timeMin, timeMax);

            if(delayMax > 0f)
            {
                delay = !delay;
                if (delay)
                {
                    timeCurrent = UnityEngine.Random.Range(delayMin, delayMax);
                    valueCurrent = valueCurrentOriginal;
                }
            }

            timeNext = Time.time + timeCurrent;

            if(callback != null && !IsDelayed())
            {
                callback();
               
            } else
            {
              
            }
        }
    }

}