﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using Random2 = System.Random;
using Random = UnityEngine.Random;
using System.Collections;

namespace SPQR
{
    public class CycleforceGeneral
    {
        MVRScript X;

        public float timer;
        public float forceTimer;
        public float flip;
        public Vector3 targetForce;
        public Vector3 targetTorque;
        public Vector3 currentForce;
        public Vector3 currentTorque;
        public float delay;

        public JSONStorableStringChooser torqueAxisJSON;

        public JSONStorableFloat periodJSON;
        public JSONStorableFloat periodRatioJSON;
        public JSONStorableFloat forceDurationJSON;
        public JSONStorableFloat forceFactorJSON;
        public JSONStorableFloat forceQuicknessJSON;
        public JSONStorableFloat torqueFactorJSON;
        public JSONStorableFloat torqueQuicknessJSON;
        public JSONStorableBool applyForceOnReturnJSON;



        public JSONStorableFloat generalForceScaleJSON;
        public JSONStorableFloat generalTorqueScaleJSON;
        public JSONStorableFloat generalLengthScaleJSON;
        public JSONStorableFloat generalForceQuicknessScaleJSON;
        public JSONStorableFloat generalTorqueQuicknessScaleJSON;
        public JSONStorableFloat generalIntensityScaleJSON;


        public List<string> receiverChoices;
        public Dictionary<string, ForceReceiver> receiverNameToForceReceiver;


        public Rigidbody RB;

        public void ResetForce()
        {
            timer = -1f;
            forceTimer = -1f;
            SetTargets(0f);
            receiverChoiceJSON = null;
        }
        public CycleforceGeneral(MVRScript script)
        {
            X = script;
        }
        public void Update()
        {
            try
            {
        
                timer -= Time.deltaTime;
                forceTimer -= Time.deltaTime;
                if (timer < 0.0f)
                {
                    if ((flip > 0f && periodRatioJSON.val != 1f) || periodRatioJSON.val == 0f)
                    {
                        if (applyForceOnReturnJSON.val)
                        {
                            flip = -1f;
                        }
                        else
                        {
                            flip = 0f;
                        }
                        timer = periodJSON.val * (1f - periodRatioJSON.val);
                        forceTimer = forceDurationJSON.val * periodJSON.val;
                    }
                    else
                    {
                        flip = 1f;
                        timer = periodJSON.val * periodRatioJSON.val;
                        forceTimer = forceDurationJSON.val * periodJSON.val;
                    }
                    SetTargets(flip);
                }
                else if (forceTimer < 0.0f)
                {
                    SetTargets(0f);
                }
            } catch (Exception e)
            {
              
            }
        }


        public void FixedUpdate(string cf)
        {
            
            // apply forces here
            float timeFactor = Time.fixedDeltaTime;
            currentForce = Vector3.Lerp(currentForce, targetForce, timeFactor * forceQuicknessJSON.val);
            currentTorque = Vector3.Lerp(currentTorque, targetTorque, timeFactor * torqueQuicknessJSON.val);
            if (RB && (!SuperController.singleton || !SuperController.singleton.freezeAnimation))
            {
                
                RB.AddForce(currentForce, ForceMode.Force);
                RB.AddTorque(currentTorque, ForceMode.Force);
            }
            
        }

        public virtual Vector3 AxisToVector(ForceProducerV2.AxisName axis)
        {
            Vector3 result;
            if (RB != null)
            {
                switch (axis)
                {
                    case ForceProducerV2.AxisName.X:
                        result = RB.transform.right;
                        break;
                    case ForceProducerV2.AxisName.NegX:
                        result = -RB.transform.right;
                        break;
                    case ForceProducerV2.AxisName.Y:
                        result = RB.transform.up;
                        break;
                    case ForceProducerV2.AxisName.NegY:
                        result = -RB.transform.up;
                        break;
                    case ForceProducerV2.AxisName.Z:
                        result = RB.transform.forward;
                        break;
                    case ForceProducerV2.AxisName.NegZ:
                        result = -RB.transform.forward;
                        break;
                    default:
                        result = Vector3.zero;
                        break;
                }
            }
            else
            {
                result = Vector3.zero;
            }
            return (result);
        }

        protected void SetTargets(float percent)
        {
            targetForce = AxisToVector(_forceAxis) * percent * forceFactorJSON.val;
            targetTorque = AxisToVector(_torqueAxis) * percent * torqueFactorJSON.val;
        }

        public void SyncReceiver(string receiver)
        {
            if (receiver != null)
            {
                ForceReceiver fr;
                if (receiverNameToForceReceiver.TryGetValue(receiver, out fr))
                {
                    RB = fr.GetComponent<Rigidbody>();
                }
                else
                {
                    RB = null;
                }
            }
            else
            {
                RB = null;
            }
        }
        public JSONStorableStringChooser receiverChoiceJSON;

        public ForceProducerV2.AxisName _forceAxis = ForceProducerV2.AxisName.Z;
        public void SyncForceAxis(string axisName)
        {
            try
            {
                ForceProducerV2.AxisName an = (ForceProducerV2.AxisName)System.Enum.Parse(typeof(ForceProducerV2.AxisName), axisName);
                _forceAxis = an;
            }
            catch (System.ArgumentException)
            {
                Debug.LogError("Attempt to set axis to " + axisName + " which is not a valid axis name");
            }
        }
        public JSONStorableStringChooser forceAxisJSON;


        public ForceProducerV2.AxisName _torqueAxis = ForceProducerV2.AxisName.X;
        public void SyncTorqueAxis(string axisName)
        {
            try
            {
                ForceProducerV2.AxisName an = (ForceProducerV2.AxisName)System.Enum.Parse(typeof(ForceProducerV2.AxisName), axisName);
                _torqueAxis = an;
            }
            catch (System.ArgumentException)
            {
                Debug.LogError("Attempt to set axis to " + axisName + " which is not a valid axis name");
            }
        }


        public void prepareRandom()
        {
            delay = (float)Random.Range(0f, 0.5f);
            periodJSON.val = (float)Random.Range(0.2f, 0.5f);
            forceFactorJSON.val = (float)Random.Range(5f, 40f);
            torqueFactorJSON.val = (float)Random.Range(5f, 10f);

            forceAxisJSON.val = SPQRUtils.RandomFromString_p67(new string[] { "X", "Y", "Z" });
            torqueAxisJSON.val = SPQRUtils.RandomFromString_p67(new string[] { "X", "Y", "Z" });

            torqueQuicknessJSON.val = (float)Random.Range(5f, 30f);
            forceQuicknessJSON.val = (float)Random.Range(5f, 30f);

        }


        public void stopForce()
        {
            forceFactorJSON.val = 0f;
            torqueFactorJSON.val = 0f;
            receiverChoiceJSON.val = null;
        }

        public bool inProgress = false;
      


        public IEnumerator CoroutineStop(float fadeDuration)
        {
          
            inProgress = true;
            float initialVal = forceFactorJSON.val;
            float targetVal = 0f;

            float initialVal2 = torqueFactorJSON.val;
            float targetVal2 = 0f;

            float elapsedTime = 0f;
            bool done = false;
          
            while (elapsedTime < fadeDuration)
            {
                
                elapsedTime += Time.deltaTime;

                if (!done && elapsedTime >= fadeDuration)
                {
                    //SuperController.LogMessage("CoroutineStop_p67 done");
                    done = true;
                    inProgress = false;

                }
                forceFactorJSON.val = Mathf.Lerp(initialVal, targetVal, elapsedTime / fadeDuration);
                torqueFactorJSON.val = Mathf.Lerp(initialVal2, targetVal2, elapsedTime / fadeDuration);

                yield return null;
            }
        }


    }
}
