/** 
 *  based on geesp0t, Doc0ctopuss spanking plugins; i believe original original credits goes to AcidBubbles
 * 
 */

using System;
using UnityEngine;
using SPQR;

namespace customevent
{
    public class EventType
    {   // using enums crashes vam
        public const byte NONE = 0;
        public const byte ENTERED = 1;
        public const byte HIT = 2;
        public const byte EXITED = 3;
        public const byte STAY = 4;
    }

    public class TriggerEventArgs : EventArgs
    {
        public Collision collision { get; set; }
        public Collider collider { get; set; }
        public byte evtType { get; set; }
        public bool insideTrigger { get; set; }
    }

    public class TriggerCollide : MonoBehaviour
    {
        string scriptkey;
        TriggerEventArgs lastEvent;
        bool clear = false;
        bool inside = false;

        public static float minTimeBetweenCollisions = 0.25f;
        public static float minTimeBetweenEnters = 0.0001f;
        float lastCollisionTime = 0;
        float lastEnterTime = 0;


        float lastScriptCheck = 0f;
        float deltaScriptCheck = 5f;

        string suffix;


        public event EventHandler<TriggerEventArgs> OnCollideAlive;

        public void setSuffix(string s)
        {
            suffix = s;
        }
        public void SetScriptkey(string s)
        {
            this.scriptkey = s;
        }

        void Awake()
        {
            lastEvent = new TriggerEventArgs
            {
                evtType = EventType.NONE,
                collider = null,
                collision = null,
                insideTrigger = false
            };
        }

        public void OnCollideAliveStop(bool stop)
        {
            clear = stop;
            //   SuperController.LogMessage("Stopped!");
        }

        private void OnTriggerEnter(Collider other)
        {
            if (clear) return;
            //other.isTrigger = false;
            inside = true;
            DoCollideEvent(EventType.ENTERED, other, null);
        }

        private void OnTriggerExit(Collider other)
        {
            inside = false;
            lastEvent.evtType = EventType.EXITED;
            DoCollideEvent(EventType.EXITED, other, null);
        }

        protected virtual void OnCollideEvent(TriggerEventArgs e)
        {
            if (clear) return;
            //SuperController.LogMessage("sending event: " + e.collider.name + ", " + e.insideTrigger);
            EventHandler<TriggerEventArgs> handler = OnCollideAlive;
            handler?.Invoke(this, e);
        }

        protected virtual void OnCollisionEnter(Collision collision)
        {
            DoCollideEvent(EventType.HIT, collision.collider, collision);
        }

        protected virtual void OnCollisionExit(Collision collision)
        {
            lastEvent.evtType = EventType.EXITED;
        }

        private void DoCollideEvent(byte evtType, Collider col, Collision c)
        {
            //   if (scriptkey_p67 != GlobalVariables.Instance.mainScriptUid_p67) return;
            //   if (!GlobalVariables.Instance.mainScriptOn_p67) return;
            //SuperController.LogMessage("collide: " + col.name + "time: " + Time.timeSinceLevelLoad);
            if (clear) return;


            if (Time.timeSinceLevelLoad - lastScriptCheck > deltaScriptCheck)
            {
                lastScriptCheck = Time.timeSinceLevelLoad;
                var id = SPQRUtils.getAtomStorableString_p67("AliveGlobalData-" + suffix, "Text", "text");
                if (id.val != scriptkey)
                {
                    Clear();
                }
                return;
            }


            bool allow = false;
            if (evtType == EventType.HIT && Time.timeSinceLevelLoad - lastCollisionTime > minTimeBetweenCollisions) allow = true;
            if (evtType == EventType.ENTERED && Time.timeSinceLevelLoad - lastEnterTime > minTimeBetweenEnters) allow = true;
            if (allow)
            {
                if (evtType == lastEvent.evtType || (evtType == EventType.HIT && c == null))
                {
                    return;
                }
                else
                {


                    //if (SPQRUtils.checkChance_p67(5f))
                    //{
                    //    SuperController.LogMessage("collide " + scriptkey_p67 + " vs : " + col.name + "time: " + Time.timeSinceLevelLoad);
                    //}

                    if (evtType == EventType.HIT) lastCollisionTime = Time.timeSinceLevelLoad;
                    if (evtType == EventType.ENTERED) lastEnterTime = Time.timeSinceLevelLoad;

                    TriggerEventArgs tempEvent = new TriggerEventArgs
                    {
                        evtType = evtType,
                        collider = col,
                        collision = c,
                        insideTrigger = inside
                    };

                    OnCollideEvent(tempEvent);
                    lastEvent = tempEvent;
                }
            }
        }

        public void Clear()
        {
            // SuperController.LogMessage("Cleared!");
            clear = true;
            if (OnCollideAlive == null) return;
            foreach (EventHandler<TriggerEventArgs> d in OnCollideAlive.GetInvocationList())
            {
                if (d == null) return;
                OnCollideAlive -= d;
            }
            OnCollideAlive = null;

        }
    }


}
